/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "fitter.hpp"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>

#include <boost/numeric/ublas/io.hpp>


namespace myfitter {

// parameter mappings
//
// pmap maps [-inf,+inf] to [0,+inv] with the asymptotic behaviour
// pmap(x) = x for x -> +inf. dpmap is the derivative of pmap and
// invpmap is the inverse function

static inline double pmap(double x)
{
    return std::max(0.5*(x + std::sqrt(x*x + 1.)),
                    std::numeric_limits<double>::epsilon());
}

static inline double dpmap(double x)
{
    double s = std::sqrt(x*x + 1.);
    if(0.5*(x+s) < std::numeric_limits<double>::epsilon()) return 0.;
    return 0.5*(1. + x/s);
}

static inline double invpmap(double x)
{
    if(x <= std::numeric_limits<double>::epsilon())
        x = std::numeric_limits<double>::epsilon();
    return x - 0.25/x;
}


// Fitter::LogLikelihoodFunction

Fitter::LogLikelihoodFunction::LogLikelihoodFunction(Fitter& fitter,
                                             Model& model,
                                             const ObservableVector& cvals,
                                             int verbosity) :
    ObjectiveFunction(),
    fitter_(&fitter),
    model_(&model),
    cvals_(&cvals)
{
    status_ = 0;
    verbose_ = verbosity;
    for(int i=0; i<model_->nparameters(); ++i)
        if(!model_->fixed(i)) freepar_.push_back(i);
    resize_(freepar_.size());
    if(ndim() == 0) {
        fitter_->calc(*model_, *cvals_);
        return;
    }
    chisq0_ = fitter_->input_function().calc(*cvals_, *cvals_, true);
    if(verbose_ >= 2) {
        std::cout << "floating parameters: ";
        for(int i=0; i<int(freepar_.size()); ++i)
            std::cout << freepar_[i] << " ";
        std::cout << "\n" << std::endl;
    }
}

void Fitter::LogLikelihoodFunction::get_minimization_parameters
(ParameterVector& x) const
{
    assert(nfreepar() == int(x.size()));

    for(int i=0; i<nfreepar(); ++i) {
        double xi=0.;
        if(!model_->has_lower_limit(freepar_[i]) &&
           !model_->has_upper_limit(freepar_[i])) {

            xi = model_->parameter(freepar_[i])/
                model_->scale(freepar_[i]);

        } else if(model_->has_lower_limit(freepar_[i]) &&
                  !model_->has_upper_limit(freepar_[i])) {

            xi = (model_->parameter(freepar_[i])
                  -model_->lower_limit(freepar_[i]))/
                model_->scale(freepar_[i]);
            xi = invpmap(xi);

        } else if(!model_->has_lower_limit(freepar_[i]) &&
                  model_->has_upper_limit(freepar_[i])) {

            xi = (model_->upper_limit(freepar_[i])-
                  model_->parameter(freepar_[i]))/
                model_->scale(freepar_[i]);
            xi = -invpmap(xi);
                
        } else if(model_->has_lower_limit(freepar_[i]) &&
                  model_->has_upper_limit(freepar_[i])) {

            double delta = (model_->upper_limit(freepar_[i]) -
                            model_->lower_limit(freepar_[i]))/
                model_->scale(freepar_[i]);
            double q = (model_->upper_limit(freepar_[i]) -
                        model_->parameter(freepar_[i]))/
                model_->scale(freepar_[i]);
            q = invpmap(q);
            xi = invpmap(invpmap(delta) - q);
        }

        x[i] = xi;
    }
}


double Fitter::LogLikelihoodFunction::operator()(const ParameterVector& x)
{
    for(int i=0; i<int(freepar_.size()); ++i) {
        double pari;
        if(!model_->has_lower_limit(freepar_[i]) &&
           !model_->has_upper_limit(freepar_[i])) {
        
            pari = x[i]*model_->scale(freepar_[i]);

        } else if(model_->has_lower_limit(freepar_[i]) &&
                  !model_->has_upper_limit(freepar_[i])) {

            double xi = x[i];
            pari = model_->lower_limit(freepar_[i]) + 
                pmap(xi)*model_->scale(freepar_[i]);

        } else if(!model_->has_lower_limit(freepar_[i]) &&
                  model_->has_upper_limit(freepar_[i])) {

            double xi = x[i];
            pari = model_->upper_limit(freepar_[i]) -
                pmap(-xi)*model_->scale(freepar_[i]);
            
        } else {

            double xi = x[i];
            double delta = (model_->upper_limit(freepar_[i]) -
                            model_->lower_limit(freepar_[i]))/
                model_->scale(freepar_[i]);
            double q = invpmap(delta) - pmap(xi);
            pari = model_->upper_limit(freepar_[i]) - 
                pmap(q)*model_->scale(freepar_[i]); 
        }

        model_->parameter(freepar_[i], pari);
    }

    if(verbose_ >= 2) {
        std::cout << "parameters: (";
        for(int i=0; i<int(freepar_.size()-1); ++i)
            std::cout << model_->parameter(freepar_[i]) << ", ";
        std::cout << model_->parameter(freepar_[freepar_.size()-1]) << ")"
                  << std::endl;
        if(verbose_ >= 3) {
            std::cout << "internal parameters: (";
            for(int i=0; i<int(freepar_.size()-1); ++i)
                std::cout << x[i] << ", ";
            std::cout << x[freepar_.size()-1] << ")"
                      << std::endl;
        }
    }

    int status=0;
    if((status=model_->calc())) throw EFailedModelEvaluation(status);
    double chisquare =
        fitter_->input_function().
        calc(model_->observables(), *cvals_, true) - chisq0_;
    if(verbose_ >= 2)
        std::cout << "chisquare: " << chisquare << std::endl;

    return chisquare;
}

void Fitter::LogLikelihoodFunction::get_fdf(const ParameterVector& x,
                                        double& f,
                                        ParameterVector& g)
{
    const IndexVector& claimedobs
        = fitter_->input_function().claimed_observables();

    f = this->operator()(x);
    int status=0;
    if((status = model_->calc_deriv())) throw EFailedModelEvaluation(status);

    ObservableVector dummy(model_->nobservables());
    fitter_->input_function().calc_deriv(model_->observables(),
                                         *cvals_, dummy);

    ObservableVector gradlfunc(claimedobs.size());
    Matrix derivs(freepar_.size(),claimedobs.size());
    for(int iobs=0; iobs<int(claimedobs.size()); ++iobs) {
        gradlfunc[iobs] = dummy[claimedobs[iobs]];
        for(int ipar=0; ipar < int(freepar_.size()); ++ipar)
            derivs(ipar,iobs)
                = model_->derivatives()(freepar_[ipar],claimedobs[iobs]);
    }
    ParameterVector gradchisq(prod(derivs, gradlfunc));
    
    for(int i=0; i<int(freepar_.size()); ++i) {
        if(!model_->has_lower_limit(freepar_[i]) &&
           !model_->has_upper_limit(freepar_[i])) {

            g[i] = gradchisq[i]*model_->scale(freepar_[i]);

        } else if(model_->has_lower_limit(freepar_[i]) &&
                  !model_->has_upper_limit(freepar_[i])) {

            double xi = x[i];
            g[i] = gradchisq[i]*dpmap(xi)*model_->scale(freepar_[i]);

        } else if(!model_->has_lower_limit(freepar_[i]) &&
                  model_->has_upper_limit(freepar_[i])) {

            double xi = x[i];
            g[i] = gradchisq[i]*dpmap(-xi)*model_->scale(freepar_[i]);
            
        } else {

            double xi = x[i];
            double delta = (model_->upper_limit(freepar_[i]) -
                            model_->lower_limit(freepar_[i]))/
                model_->scale(freepar_[i]);
            double q = delta - 0.25/delta - pmap(xi);
            g[i] = gradchisq[i]*dpmap(q)*dpmap(xi)*model_->scale(freepar_[i]);
        }
    }

    if(verbose_ >= 4) {
        std::cout << "gradient: (";
        for(int i=0; i<int(freepar_.size()-1); ++i)
            std::cout << gradchisq[i] << ", ";
        std::cout << gradchisq[freepar_.size()-1] << ")"
                  << std::endl;
        std::cout << "internal gradient: (";
        for(int i=0; i<int(freepar_.size()-1); ++i)
            std::cout << g[i] << ", ";
        std::cout << g[freepar_.size()-1] << ")"
                  << std::endl;
        std::cout << std::endl;
    }
}

void Fitter::LogLikelihoodFunction::get_gradient(const ParameterVector& x,
                                             ParameterVector& g)
{
    double f;
    get_fdf(x, f, g);
}


// class Fitter

void Fitter::get_starting_point(Model& model, const ObservableVector& cvals)
{
    if(model.nsamples() <= 0) return;
    int besti = 0;
    double chi2min = ifunc_.calc(model.sample_observables(0), cvals, true);
    double chi2 = 0.;
    for(int i=1; i<model.nsamples(); ++i) {
        chi2 = ifunc_.calc(model.sample_observables(i), cvals, true);
        if(chi2 < chi2min) {
            chi2min = chi2;
            besti = i;
        }
    }
    model.parameters(model.sample_parameters(besti));
}

int Fitter::local_fit(Model& model, const ObservableVector& cvals)
{
    int result = local_fit_(model, cvals);
    if(result == GSL_SUCCESS)
        return calc(model, cvals);
    else {
        calc(model, cvals);
        return result;
    }
}

int Fitter::local_fit_(Model& model, const ObservableVector& cvals)
{
    // initialize chi square function
    LogLikelihoodFunction fcn;
    get_log_likelihood_function(fcn, model, cvals);

    // initialize minimizer
    if(minimizer_->ndim() != fcn.nfreepar())
        minimizer_->resize(fcn.nfreepar());
    minimizer_->verbosity(minimizer_verbosity());
    minimizer_->precision(minimizer_precision());
    minimizer_->line_search_precision(minimizer_line_search_precision());
    minimizer_->maxiter(minimizer_iterations());
    minimizer_->keep_hessian(minimizer_keep_hessian());

    ParameterVector x(fcn.nfreepar());
    fcn.get_minimization_parameters(x);
    minimizer_->x(x);

    // minimize
    int status = minimizer_->minimize(fcn);
    if(status != GSL_SUCCESS && minimizer_->verbosity() > 0)
        std::cout << "minimization failed with exit code "
                  << status << "\nat parameters: "
                  << model.parameters()
                  << std::endl;
    return status;
}


void Fitter::ScanIntegrand::operator()(const double x[], const int k[],
                                       const double& weight,
                                       const double aux[],
                                       double f[])
{
    ParameterVector smin(mod_->nparameters());
    ParameterVector smax(mod_->nparameters());
    for(int i=0; i<mod_->nparameters(); ++i) {
        smin[i] = mod_->scan_min(i);
        smax[i] = mod_->scan_max(i);
        if(mod_->has_lower_limit(i))
            smin[i] = std::max(smin[i], mod_->lower_limit(i));
        if(mod_->has_upper_limit(i))
            smax[i] = std::min(smax[i], mod_->upper_limit(i));
    }
    
    for(int j=0; j<mod_->nparameters(); ++j) {
        if(!mod_->fixed(j)) {
            mod_->parameter(j, (1.-x[j])*smin[j] + x[j]*smax[j]);
        }
    }
    mod_->calc();
    double chisq =
        ifunc_->calc(mod_->observables(),cvals_, true)
        -ifunc_->calc(cvals_,cvals_, true);
    f[0] = 1./std::pow(chisquare_offset_ + chisq, chisquare_power_);
    if(store_) mod_->store();
}


void Fitter::adaptive_scan(Model& model, const ObservableVector& cvals,
                           HepSource::Int64 nfirstshots,
                           HepSource::Int64 nshots, int niter,
                           HepSource::Int64 nlastshots)
{
    ScanIntegrand s(get_scan_integrand(model, cvals));
    s.store(false);

    int verbose = dvegas_verbosity();

    HepSource::Dvegas::CellAccumulators ca;
    HepSource::Dvegas dv(scans_nbins_, s);
    ca = dv.collectData(nfirstshots,verbose);
    for(int i=0; i<niter; ++i) {
        dv.adapt(ca);
        if(verbose > 0) dv.info();
        ca = dv.collectData(nshots,verbose);
    }
    dv.adapt(ca);
    if(verbose > 0) dv.info();

    s.store(true);
    ca = dv.collectData(nlastshots,verbose);
}


}
