/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "fitter.hpp"
#include <limits>

#include "ublasutils/eigen.hpp"
#include "ublasutils/orthonormalize.hpp"
#include "ublasutils/determinant.hpp"

#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/math/special_functions/erf.hpp>

#include <dvegas/omnicomp.h>


namespace myfitter {

typedef boost::numeric::ublas::matrix_column<Matrix> MatrixColumn;
typedef boost::numeric::ublas::matrix_row<Matrix> MatrixRow;
typedef boost::numeric::ublas::identity_matrix<double> IdentityMatrix;
typedef boost::numeric::ublas::zero_vector<double> ZeroVector;
using boost::numeric::ublas::prod;
using boost::numeric::ublas::inner_prod;
using boost::numeric::ublas::trans;
using boost::numeric::ublas::matrix_range;
using boost::numeric::ublas::range;
using boost::numeric::ublas::identity_matrix;
using boost::numeric::ublas::row;
using boost::numeric::ublas::column;

using ublasutils::orthonormalize;
using ublasutils::jacobi_diag_symm;
using ublasutils::det;


// class Fitter::PValueIntegrand

Fitter::PValueIntegrand::PValueIntegrand
(Model& fmod, Model& cmod, Fitter& fitter, const IndexVector& claimedobs,
 const Matrix& lam2obs, const ObservableVector& origin,
 const ObservableVector& lambda0, int n0, int n1, double deltachisq,
 double jacobian, int verbose, bool adaptive) : 
    HepSource::Integrand(), fmod_(&fmod), cmod_(&cmod),
    fitter_(&fitter), claimedobs_(claimedobs), lam2obs_(lam2obs),
    origin_(origin), lambda0_(lambda0), ndim0_(n0), ndim1_(n1),
    ndim_(claimedobs.size()), nvar_(ndim_+1), dchisq_(deltachisq),
    jacobian_(jacobian), nfailedshots_(0), verbose_(verbose),
    adaptive_(adaptive) { }

void Fitter::PValueIntegrand::set_lambda_(const double x[], const double y[],
                                          int ndim, int offset, double deltasq,
                                          ObservableVector& lambda, 
                                          double& jacobian)
{
    if(ndim <= 0) return;

    double p0=boost::math::gamma_p(0.5*ndim,0.5*deltasq);
    double alpha=fitter_->inner_region_power();
    double f=fitter_->inner_region_fraction();
    double delta = std::sqrt(deltasq);
    double rsq, r;
    if(deltasq > 0.) {
        // pick radius

        if(x[0] < f) {
            rsq = deltasq*std::pow(x[0]/f, 2./(ndim+alpha));
            r = std::sqrt(rsq);
            jacobian *= 2.*std::pow(M_PI,0.5*ndim)*std::pow(delta, ndim+alpha)/
                (f*(ndim+alpha)*boost::math::tgamma(0.5*ndim)
                 *std::pow(r,alpha));
        } else {
            double p = (x[0]-f)/(1.-f);
            rsq = 2.*boost::math::gamma_p_inv(0.5*ndim,(1.-p0)*p + p0);
            if(!(rsq <= (std::numeric_limits<double>::max())
                 && rsq >= -std::numeric_limits<double>::max()))
                throw ELargeDeltaChiSquare();
            r = std::sqrt(rsq);
            jacobian *= std::pow(2.*M_PI, 0.5*ndim)*(1.-p0)/
                ((1.-f)*std::exp(-0.5*rsq));
        }

        // construct lambda vector
        double rinv=0.;
        for(int i=0; i<ndim; ++i)
            rinv += y[i]*y[i];
        if(rinv != 0.)
            rinv = 1./std::sqrt(rinv);
        else
            rinv = rsq = 0.;
        for(int i=0; i<ndim; ++i)
            lambda[offset+i] = rinv*y[i]*r;
    } else {
        rsq = 0.;
        for(int i=0; i<ndim; ++i) {
            lambda[offset+i] = y[i];
            rsq += y[i]*y[i];
        }
        jacobian *= std::pow(2.*M_PI, 0.5*ndim)*std::exp(0.5*rsq);
    }
}

void Fitter::PValueIntegrand::operator()(const double x[], const int k[],
                                         const double& weight,
                                         const double aux[], double f[])
{
    if(!adaptive_) x = aux;

    // construct ndim_ normally distributed random variables
    double y[ndim_];
    for(int i=0; i<ndim_; ++i)
        y[i] = M_SQRT2*boost::math::erf_inv(2.*x[i+1]-1.);

    // construct lambda vector
    double jacobian = jacobian_;
    ObservableVector lambda(ndim_);
    set_lambda_(x, y, ndim0_, 0, 0., lambda, jacobian);
    double rsq = 0.;
    for(int i=0; i<ndim0_; ++i)
        rsq += std::pow(lambda[i]-lambda0_[i], 2);
    set_lambda_(x, y+ndim0_, ndim1_-ndim0_, ndim0_, rsq+dchisq_,
                lambda, jacobian);
    set_lambda_(x+1, y+ndim1_, ndim_-ndim1_, ndim1_, 0.,
                lambda, jacobian);

    // calculate delta chisquare and true likelihood
    ObservableVector dummy = prod(lam2obs_, lambda) + origin_;
    ObservableVector toyobs(fmod_->nobservables());
    ObservableVector obs0(fmod_->nobservables());
    for(int i=0; i < int(toyobs.size()); ++i) {
        toyobs[i] = Model::nan();
        obs0[i] = Model::nan();
    }
    for(int i=0; i < ndim_; ++i) {
        toyobs[claimedobs_[i]] = dummy[i];
        obs0[claimedobs_[i]] = origin_[i];
    }
    if(verbose_ >= 3) std::cout << "toy observables: " << toyobs << std::endl;
    int status;
    if(verbose_ >= 3) std::cout << "fitting full model: ";
    if((status = fitter_->global_fit(*fmod_, toyobs))) {
        std::cerr << "\
fit of full model returned status " << status << " for these toy measurements:\n";
        for(int i=0; i<fitter_->nobservables(); ++i)
            std::cerr << toyobs[i] << " ";
        std::cerr << std::endl;
        nfailedshots_ += 1;
    }
    if(verbose_ >= 3) std::cout << fmod_->chisquare() << std::endl;
    if(verbose_ >= 3) std::cout << "fitting constrained model: ";
    if((status = fitter_->global_fit(*cmod_, toyobs))) {
        std::cerr << "\
fit of constrained model returned status " << status << " for these toy measurements:\n";
        for(int i=0; i<fitter_->nobservables(); ++i)
            std::cerr << toyobs[i] << " ";
        std::cerr << std::endl;
        nfailedshots_ += 1;
    }
    if(verbose_ >= 3) std::cout << cmod_->chisquare() << std::endl;
    double dchisq = cmod_->chisquare() - fmod_->chisquare();
    double lval = fitter_->input_function().
        calc_nosystematics(obs0, toyobs);

    if(std::isnan(dchisq)) dchisq = 0.;
    if(dchisq > dchisq_)
        f[0] = std::exp(-0.5*lval)*jacobian;
    else
        f[0] = 0.;

    // rsq = 0.;
    // for(int i=0; i<ndim_; ++i)
    //     rsq += lambda[i]*lambda[i];
    // f[0] = std::pow(2.*M_PI,-0.5*ndim_)*std::exp(-0.5*rsq)*jacobian;

    if(verbose_ >= 3) std::cout << "integrand value: " << f[0] << std::endl;
}


// class Fitter

void Fitter::assign_deriv_(Matrix& m, int icol, const Model& mod, int ipar)
{
    int nobs = input_function().n_claimed_observables();
    ObservableVector vderiv(nobs);
    for(int j=0; j < nobs; ++j)
        vderiv[j] = mod.derivatives()(ipar, input_function().
                                      claimed_observables()[j]);
    MatrixColumn col(m, icol);
    col.assign(prod(qo_, vderiv*mod.scale(ipar)));
}

void Fitter::set_conversion_matrices_(const Model& fmod, const Model& cmod)
{
    int nobs = input_function().n_claimed_observables();
    cobs_.resize(nobs,false);
    fobs_.resize(nobs,false);
    qo_.resize(nobs,nobs,false);
    oq_.resize(nobs,nobs,false);

    for(int i=0; i < nobs; ++i) {
        cobs_[i] = cmod.
            observables()[input_function().claimed_observables()[i]];
        fobs_[i] = fmod.
            observables()[input_function().claimed_observables()[i]];
    }
    Matrix hessian(nobservables(),nobservables());
    Matrix vinv(nobs, nobs);
    ifunc_.get_hessian(hessian, cmod.observables());
    for(int i=0; i < nobs; ++i)
        for(int j=0; j < nobs; ++j)
            vinv(i,j) = hessian(input_function().claimed_observables()[i],
                                input_function().claimed_observables()[j]);

    Matrix umat = IdentityMatrix(nobs);
    jacobi_diag_symm(vinv, umat);
    for(int i=0; i<nobs; ++i) {
        double sqrtvinvi = std::sqrt(vinv(i,i));
        MatrixRow qorow(qo_, i);
        qorow = sqrtvinvi*MatrixColumn(umat, i);
        MatrixColumn oqcol(oq_, i);
        oqcol = MatrixColumn(umat, i)/sqrtvinvi;
    }
}

double Fitter::calc_nested_lrt_pvalue(Model& fmod, Model& cmod)
{
    assert(fmod.nparameters() == cmod.nparameters());
    assert(fmod.nobservables() == nobservables());
    assert(cmod.nobservables() == nobservables());

    // save delta chisquare
    double dchisq = cmod.chisquare() - fmod.chisquare();

    // conversion to and from normalized observables
    int nobs = input_function().n_claimed_observables();
    set_conversion_matrices_(fmod,cmod);
    // qobs = qo_*(obs-cobs_)
    // obs = oq_*qobs + cobs_


    // construct basis vmat in the space of normalized observables
    // with the first n1 basis vectors spanning the constrained
    // model's hyperplane and the first n2 vectors spanning the
    // full model's hyperplane

    ParameterVector fmodpar = fmod.parameters();
    ParameterVector cmodpar = cmod.parameters();

    IndexVector parind1; // parameters floating in both models
    IndexVector parind2; // parameters floating only in the full model
    for(int i=0; i < fmod.nparameters(); ++i) {
        if(!cmod.fixed(i)) {
            assert(!fmod.fixed(i));
            parind1.push_back(i);
        } else if(!fmod.fixed(i))
            parind2.push_back(i);
    }

    Matrix vmat(nobs, parind1.size()+parind2.size()+nobs);
    assert(fmod.calc_deriv() == 0);

    // construct subspace of constrained model in vmat(i,0..n1-1)
    for(int i=0; i < int(parind1.size()); ++i)
        assign_deriv_(vmat, i, fmod, parind1[i]);
    int n1 = orthonormalize(vmat, orthogonalization_tolerance(),
                            0, parind1.size());
    
    // construct complement of constrained model in full model
    // in vmat(i,n1..n2-1)
    fmod.parameters(cmod.parameters());
    for(int i=0; i < int(parind2.size()); ++i)
        assign_deriv_(vmat, n1+i, fmod, parind2[i]);
    int n2 = orthonormalize(vmat,orthogonalization_tolerance(),
                            n1,n1+parind2.size());

    // construct complement of full model in vmat(i,n2..nobservables()-1)
    {
        matrix_range<Matrix> vmatrange(vmat,
                                       range(0, nobs),
                                       range(n2, n2+nobs));
        vmatrange.assign(identity_matrix<double>(nobs));
    }
    orthonormalize(vmat, orthogonalization_tolerance(), n2);

    // re-order columns and assign to umat
    Matrix umat(nobs,nobs);
    for(int i=n1; i<n2; ++i) {
        MatrixColumn umatcol(umat, i-n1);
        umatcol.assign(MatrixColumn(vmat, i));
    }
    for(int i=0; i<n1; ++i) {
        MatrixColumn umatcol(umat, n2-n1+i);
        umatcol.assign(MatrixColumn(vmat, i));
    }
    for(int i=n2; i<nobs; ++i) {
        MatrixColumn umatcol(umat, i);
        umatcol.assign(MatrixColumn(vmat, i));
    }

    // vmat = oq*umat
    vmat.resize(nobs,nobs,false);
    vmat.assign(prod(oq_, umat));


    // perform DVegas integration

    ObservableVector lam0(nobs);
    for(int i=0; i<nobs; ++i) lam0[i] = Model::nan();
    PValueIntegrand integrand(fmod, cmod, *this, 
                              input_function().claimed_observables(),
                              vmat, cobs_, lam0, 0, n2-n1, dchisq,
                              std::abs(1./det(qo_)), dvegas_verbosity(),
                              dvegas_adaptive());

    int cdim=integrand.nvar();
    int adim=0;
    if(!dvegas_adaptive()) {
        cdim = 0;
        adim = integrand.nvar();
    }
    HepSource::Dvegas dv(cdim,               // continuous dimensions
                         dvegas_nbins(),     // number of bins
                         std::vector<int>(), // no correlated dimensions
                         std::vector<int>(), // no discrete dimensions
                         adim,               // number of auxiliary dimensions
                         1,                  // number of integrands
                         integrand);         // integrand object
    HepSource::Dvegas::CellAccumulators ca;
    nshots_ = 0;
    ca = dv.collectData(dvegas_nfirstshots(),dvegas_verbosity());
    nshots_ += dvegas_nfirstshots();
    for(int i=0; i<dvegas_niterations(); ++i) {
        dv.adapt(ca);
        if(dvegas_verbosity() > 0) dv.info();
        if(std::abs(dv.stats(0).standardDeviation()/
                    dv.stats(0).integral()) < dvegas_precision())
            break;
        ca = dv.collectData(dvegas_nshots(),dvegas_verbosity());
        nshots_ += dvegas_nshots();
    }
    pvalue_ = dv.stats(0).integral();
    pvalue_error_ = std::abs(dv.stats(0).standardDeviation());
    reached_precision_goal_ =
        (std::abs(pvalue_error_/pvalue_) < dvegas_precision_);
    failed_shot_ratio_ = double(integrand.failed_shots())/nshots_;

    fmod.parameters(fmodpar); fmod.calc();
    cmod.parameters(cmodpar); cmod.calc();

    return pvalue_;
}


double Fitter::calc_lrt_pvalue(Model& mod0, Model& mod1, int argc, char** argv)
{
    bool with_omnicomp = (argc != 0 && argv != 0);

    assert(mod0.nobservables() == nobservables());
    assert(mod1.nobservables() == nobservables());

    // save parameters and delta chisquare
    ParameterVector mod0par = mod0.parameters();
    ParameterVector mod1par = mod1.parameters();
    Model* mod[2];
    if(mod0.chisquare() < mod1.chisquare()) {
        mod[0] = &mod0;
        mod[1] = &mod1;
    } else {
        mod[0] = &mod1;
        mod[1] = &mod0;
    }
    double dchisq = mod[1]->chisquare() - mod[0]->chisquare();
    mod[0]->calc_deriv();
    mod[1]->calc_deriv();

    // conversion to and from normalized observables
    int nobs = input_function().n_claimed_observables();
    set_conversion_matrices_(*(mod[0]),*(mod[1]));
    // qobs = qo_*(obs-cobs_)
    // obs = oq_*qobs + cobs_

    // construct vector in Q-space pointing from mod[1] best-fit point
    // to mod[0] best-fit point
    ObservableVector q10(prod(qo_, fobs_ - cobs_));

    IndexVector parind0; // parameters floating in mod[0]
    IndexVector parind1; // parameters floating in mod[1]
    for(int i=0; i < mod[0]->nparameters(); ++i)
        if(!mod[0]->fixed(i) && !mod[0]->smallrange(i)) parind0.push_back(i);
    for(int i=0; i < mod[1]->nparameters(); ++i)
        if(!mod[1]->fixed(i) && !mod[1]->smallrange(i)) parind1.push_back(i);

    assert(int(parind0.size()) <= nobs);
    assert(int(parind1.size()) <= nobs);

    // construct mod[1] subspace in vmat1
    Matrix vmat1(nobs,nobs,0.);
    for(int i=0; i < int(parind1.size()); ++i)
        assign_deriv_(vmat1, i, *(mod[1]), parind1[i]);
    int n1 = orthonormalize(vmat1, orthogonalization_tolerance(),
                            0, parind1.size());

    // construct mod[0] subspace in vmat0
    Matrix vmat0(nobs,nobs+parind1.size()+parind0.size(),0.);
    for(int i=0; i < int(parind0.size()); ++i)
        assign_deriv_(vmat0, i, *(mod[0]), parind0[i]);

    // construct projection of mod[0] subspace onto mod[1] subspace
    // in vmat0(i,0..n0)
    {
        Matrix pmat(nobs,nobs);
        matrix_range<Matrix> vmat1range(vmat1, range(0, nobs), range(0, n1));
        pmat.assign(prod(vmat1range,trans(vmat1range)));
        matrix_range<Matrix> vmat0range(vmat0, range(0, nobs),
                                        range(0, parind0.size()));
        vmat0range = prod(pmat,vmat0range);
    }
    int n0 = orthonormalize(vmat0, orthogonalization_tolerance(),
                            0, parind0.size());

    // construct complement of mod[0] subspace in mod[1] subspace
    // in vmat0(i,n0..n1)
    {
        matrix_range<Matrix> vmat0range(vmat0, range(0, nobs), range(n0,n0+n1));
        matrix_range<Matrix> vmat1range(vmat1, range(0, nobs), range(0,n1));
        vmat0range.assign(vmat1range);
    }
    n1 = orthonormalize(vmat0, orthogonalization_tolerance(),
                        n0, n0+n1);

    // construct complement of mod[1] subspace in vmat0(i,n1..nobs)
    {
        matrix_range<Matrix> vmat0range(vmat0,range(0,nobs),range(n1,n1+nobs));
        vmat0range.assign(identity_matrix<double>(nobs));
    }
    orthonormalize(vmat0, orthogonalization_tolerance(), n1, n1+nobs);

    // re-order columns and assign to vmat1
    {
        matrix_range<Matrix> vmat0range(vmat0,range(0,nobs),range(n0,nobs));
        matrix_range<Matrix> vmat1range(vmat1,range(0,nobs),range(0,nobs-n0));
        vmat1range.assign(vmat0range);
    }
    {
        matrix_range<Matrix> vmat0range(vmat0,range(0,nobs),range(0,n0));
        matrix_range<Matrix> vmat1range(vmat1,range(0,nobs),
                                        range(nobs-n0,nobs));
        vmat1range.assign(vmat0range);
    }

    // vmat0 = oq_*umat
    vmat0.resize(nobs,nobs,false);
    vmat0.assign(prod(oq_, vmat1));
    
    // construct vector in lambda-space pointing from mod[1] best-fit point
    // to mod[0] best-fit point
    ObservableVector lam0(prod(trans(vmat0), q10));

    // perform DVegas integration

    int ndim0=0, ndim1=0;
    if(inner_region_fraction() >= 0.) {
        ndim0 = n1-n0;
        ndim1 = nobs-n0;
    }
    PValueIntegrand integrand(*(mod[0]), *(mod[1]), *this,
                              input_function().claimed_observables(),
                              vmat0, cobs_, lam0, ndim0, ndim1, dchisq,
                              std::abs(1./det(qo_)), dvegas_verbosity(),
                              dvegas_adaptive());

    HepSource::Dvegas::CellAccumulators ca;
    nshots_ = 0;

    HepSource::OmniComp* oc = 0;
    if(with_omnicomp) {
        HepSource::OmniComp::registerIntegrand(&integrand,
                                               "relative_pvalue_integrand");
        oc = new HepSource::OmniComp(argc, argv);
    }

    int cdim=integrand.nvar();
    int adim=0;
    if(!dvegas_adaptive()) {
        cdim = 0;
        adim = integrand.nvar();
    }
    HepSource::Dvegas dv(cdim,               // continuous dimensions
                         dvegas_nbins(),     // number of bins
                         std::vector<int>(), // no correlated dimensions
                         std::vector<int>(), // no discrete dimensions
                         adim,               // number of auxiliary dimensions
                         1,                  // number of integrands
                         integrand);         // integrand object

    if(with_omnicomp)
        dv.attachOmniComp(oc);

    ca = dv.collectData(dvegas_nfirstshots(),dvegas_verbosity());
    nshots_ += dvegas_nfirstshots();
    for(int i=0; i<dvegas_niterations(); ++i) {
        dv.adapt(ca);
        if(dvegas_verbosity() > 0) dv.info();
        if(std::abs(dv.stats(0).standardDeviation()/
                    dv.stats(0).integral()) < dvegas_precision())
            break;
        ca = dv.collectData(dvegas_nshots(),dvegas_verbosity());
        nshots_ += dvegas_nshots();
    }
    pvalue_ = dv.stats(0).integral();
    pvalue_error_ = std::abs(dv.stats(0).standardDeviation());
    reached_precision_goal_ =
        (std::abs(pvalue_error_/pvalue_) < dvegas_precision_);
    if(with_omnicomp)
        failed_shot_ratio_ = -1.;
    else
        failed_shot_ratio_ = double(integrand.failed_shots())/nshots_;

    if(with_omnicomp)
        delete oc;

    mod0.parameters(mod0par); mod0.calc();
    mod1.parameters(mod1par); mod1.calc();

    return pvalue_;
}


}
