/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "inputcomponents.hpp"

#include "ublasutils/invert.hpp"
#include "ublasutils/determinant.hpp"


namespace myfitter {

using ublasutils::invert;
using boost::numeric::ublas::prod;
using boost::numeric::ublas::inner_prod;


// class GaussianIC

double GaussianIC::calc(const ObservableVector& pred,
                        const ObservableVector& meas)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());
        
    double d = pred[0]-meas[0];
    if(std::isnan(d))
        return d;
    else if(d < -syst_minus_)
        return std::pow((d + syst_minus_)/stat_err_, 2)
            + std::log(stat_err_*stat_err_*2.*M_PI);
    else if(d > syst_plus_)
        return std::pow((d - syst_plus_)/stat_err_, 2)
            + std::log(stat_err_*stat_err_*2.*M_PI);
    else
        return std::log(stat_err_*stat_err_*2.*M_PI);
}

int GaussianIC::calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());
        
    double d = pred[0] - meas[0];
    if(std::isnan(d))
        return d;
    else if(d < -syst_minus_)
        result[0] = 2.*(d + syst_minus_)/(stat_err_*stat_err_);
    else if(d > syst_plus_)
        result[0] = 2.*(d - syst_plus_)/(stat_err_*stat_err_);
    else
        result[0] = 0.;
    return 0;
}

double GaussianIC::calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());
        
    double d = pred[0] - meas[0];
    return std::pow(d/stat_err_, 2) + std::log(stat_err_*stat_err_*2.*M_PI);
}


// class AsymmetricGaussianIC

double AsymmetricGaussianIC::calc(const ObservableVector& pred,
                                  const ObservableVector& meas)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());
        
    double stat_err = 0.5*(stat_plus_ + stat_minus_);

    double d = pred[0] - meas[0];
    if(std::isnan(d))
        return d;
    else if(d < -syst_minus_)
        return std::pow((d + syst_minus_)/stat_minus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else if(d > syst_plus_)
        return std::pow((d - syst_plus_)/stat_plus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else
        return std::log(stat_err*stat_err*2.*M_PI);
}

int AsymmetricGaussianIC::calc_deriv(const ObservableVector& pred,
                                     const ObservableVector& meas,
                                     ObservableVector& result)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());
        
    double d = pred[0] - meas[0];
    if(std::isnan(d))
        return d;
    else if(d < -syst_minus_)
        result[0]
            = 2.*(d + syst_minus_)/(stat_minus_*stat_minus_);
    else if(d > syst_plus_)
        result[0]
            = 2.*(d - syst_plus_)/(stat_plus_*stat_plus_);
    else
        result[0] = 0.;
    return 0;
}

double AsymmetricGaussianIC::calc_nosystematics(const ObservableVector& pred,
                                                const ObservableVector& meas)
{
    assert(int(pred.size()) >= n_claimed_observables());
    assert(int(meas.size()) >= n_claimed_observables());

    double stat_err = 0.5*(stat_plus_ + stat_minus_);
        
    double d = pred[0] - meas[0];
    if(d < 0.)
        return std::pow(d/stat_minus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else
        return std::pow(d/stat_plus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
}


// class CorrelatedGaussianIC

void CorrelatedGaussianIC::add(int iobs, double val, double err)
{
    claim_(iobs, val);
    int n = n_claimed_observables();
    err_.resize(n);
    err_[n-1] = err;
    cor_.resize(n, n);
    for(int i=0; i<n-2; ++i)
        cor_(i, n-1) = cor_(n-1, i) = 0.;
    cor_(n-1, n-1) = 1.;
    invvar_.resize(n, n);
    needs_update_ = true;
}

void CorrelatedGaussianIC::update_()
{
    int n = n_claimed_observables();
    Matrix var(n, n);
    for(int i=0; i<n; ++i)
        for(int j=i; j<n; ++j)
            var(i,j) = var(j,i) = cor_(i,j)*err_[i]*err_[j];

    invert(var, invvar_);
    needs_update_ = false;
}

double CorrelatedGaussianIC::calc(const ObservableVector& pred,
                                  const ObservableVector& meas)
{
    if(needs_update_) update_();
    int n = n_claimed_observables();
    boost::numeric::ublas::vector<double> d(n);
    for(int i=0; i<n; ++i)
        d(i) = pred[i] - meas[i];
    boost::numeric::ublas::vector<double> invvd(prod(invvar_, d));
    double detinvvar = myfitter::ublasutils::det(invvar_);
    return inner_prod(d, invvd) + std::log(std::pow(2.*M_PI, n)/detinvvar);
}

int CorrelatedGaussianIC::calc_deriv(const ObservableVector& pred,
                                     const ObservableVector& meas,
                                     ObservableVector& result)
{
    if(needs_update_) update_();
    int n = n_claimed_observables();
    boost::numeric::ublas::vector<double> d(n);
    for(int i=0; i<n; ++i)
        d(i) = pred[i] - meas[i];
    boost::numeric::ublas::vector<double> invvd(prod(invvar_, d));
    for(int i=0; i<n; ++i)
        result[i] = 2.*invvd[i];
    return 0;
}

void CorrelatedGaussianIC::get_hessian(Matrix& hess,
                                       const ObservableVector& obs)
{
    if(needs_update_) update_();
    hess = invvar_;
}


}
