/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "inputfunction.hpp"

typedef boost::numeric::ublas::zero_vector<double> ZeroVector;


namespace myfitter {

// class InputFunction

int InputFunction::add(const InputComponent& lc)
{
    assert(lc.nobservables() <= nobservables());
    // for(IndexVector::const_iterator iobs = lc.claimed_observables().begin();
    //     iobs != lc.claimed_observables().end(); ++iobs) {
    //     for(InputComponentVector::const_iterator ilcv
    //             = lcv_.begin(); ilcv != lcv_.end(); ++ilcv) {
    //         if(std::find((*ilcv)->claimed_observables().begin(),
    //                      (*ilcv)->claimed_observables().end(), *iobs)
    //            !=(*ilcv)->claimed_observables().end()) return -1;
    //     }
    // }
    lcv_.push_back(lc.copy());
    is_constraint_.push_back(false);

    for(IndexVector::const_iterator inew = lc.claimed_observables().begin();
        inew != lc.claimed_observables().end(); ++inew) {  
        IndexVector::iterator iold = claimed_observables_.begin();
        while(iold != claimed_observables_.end() && (*iold < *inew)) ++iold;
        if((iold == claimed_observables().end()) || *iold > *inew)
            claimed_observables_.insert(iold, *inew);
    }
    
    for(int i=0; i<lc.n_claimed_observables(); ++i)
        central_values_[lc.external_index(i)] = lc.central_values()[i];

    return lcv_.size()-1;
}

double InputFunction::calc_component(int i,
                                     const ObservableVector& pred,
                                     const ObservableVector& meas)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    assert(i >= 0 && i < ncomponents());

    ObservableVector ipred(lcv_[i]->n_claimed_observables());
    ObservableVector imeas(lcv_[i]->n_claimed_observables());
    for(int j=0; j<int(ipred.size()); ++j) {
        ipred[j] = pred[lcv_[i]->external_index(j)];
        imeas[j] = meas[lcv_[i]->external_index(j)];
    }
    return lcv_[i]->calc(ipred, imeas);
}

double InputFunction::calc(const ObservableVector& pred,
                           const ObservableVector& meas,
                           bool with_constraints)
{
    double chi2 = 0.;
    for(int i=0; i<int(lcv_.size()); ++i) {
        if(is_constraint_[i]) {
            if(with_constraints)
                chi2 += calc_component(i, pred, meas)
                    *constraint_penalty_factor_;
        } else
            chi2 += calc_component(i, pred, meas);
    }
    return chi2;
}


double InputFunction::calc_contrib(const IndexVector& iv,
                                   const ObservableVector& pred,
                                   const ObservableVector& meas,
                                   bool with_constraints)
{
    double chi2 = 0.;
    for(int i=0; i<int(lcv_.size()); ++i) {
        bool relevant = false;
        for(IndexVector::const_iterator p
                = lcv_[i]->claimed_observables().begin();
            p != lcv_[i]->claimed_observables().end();
            ++p) {
            if(std::find(iv.begin(), iv.end(), *p) != iv.end()) {
                relevant = true;
                break;
            }
        }
        if(!relevant) continue;

        if(is_constraint_[i]) {
            if(with_constraints)
                chi2 += calc_component(i, pred, meas)
                    *constraint_penalty_factor_;
        } else
            chi2 += calc_component(i, pred, meas);
    }
    return chi2;    
}


double InputFunction::calc_constraint_penalty(const ObservableVector& pred,
                                              const ObservableVector& meas)
{
    double chi2 = 0.;
    for(int i=0; i<int(lcv_.size()); ++i) {
        if(is_constraint_[i]) {
            chi2 += calc_component(i, pred, meas);
        }
    }
    return chi2;
}


double InputFunction::calc_penalty_contrib(const IndexVector& iv,
                                           const ObservableVector& pred,
                                           const ObservableVector& meas)
{
    double chi2 = 0.;
    for(int i=0; i<int(lcv_.size()); ++i) {
        if(is_constraint_[i]) {
            bool relevant = false;
            for(IndexVector::const_iterator p
                    = lcv_[i]->claimed_observables().begin();
                p != lcv_[i]->claimed_observables().end();
                ++p) {
                if(std::find(iv.begin(), iv.end(), *p) != iv.end()) {
                    relevant = true;
                    break;
                }
            }

            if(relevant)
                chi2 += calc_component(i, pred, meas);
        }
    }
    return chi2;
}


int InputFunction::calc_deriv(const ObservableVector& pred,
                              const ObservableVector& meas,
                              ObservableVector& result)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    assert(int(result.size()) == nobservables());
    result = ZeroVector(nobservables());
    int status;
    for(int i=0; i<int(lcv_.size()); ++i) {
        ObservableVector deriv(lcv_[i]->n_claimed_observables(), Model::nan());
        ObservableVector ipred(lcv_[i]->n_claimed_observables());
        ObservableVector imeas(lcv_[i]->n_claimed_observables());
            for(int j=0; j<int(ipred.size()); ++j) {
                ipred[j] = pred[lcv_[i]->external_index(j)];
                imeas[j] = meas[lcv_[i]->external_index(j)];
            }

        double cpfactor = 1.;
        if(is_constraint_[i]) cpfactor = constraint_penalty_factor_;

        if((status=lcv_[i]->calc_deriv(ipred, imeas, deriv))) return status;

        for(int j=0; j < lcv_[i]->n_claimed_observables(); ++j)
            result[lcv_[i]->external_index(j)] += deriv[j]*cpfactor;
    }
    return 0;
}

double InputFunction::calc_nosystematics(const ObservableVector& pred,
                                         const ObservableVector& meas,
                                         bool with_constraints)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    double chi2 = 0.;
    for(int i=0; i<int(lcv_.size()); ++i) {
        ObservableVector ipred(lcv_[i]->n_claimed_observables());
        ObservableVector imeas(lcv_[i]->n_claimed_observables());
        for(int j=0; j<int(ipred.size()); ++j) {
            ipred[j] = pred[lcv_[i]->external_index(j)];
            imeas[j] = meas[lcv_[i]->external_index(j)];
        }
        if(is_constraint_[i]) {
            if(with_constraints)
                chi2 += lcv_[i]->calc_nosystematics(ipred, imeas)*
                    constraint_penalty_factor_;
        } else
            chi2 += lcv_[i]->calc_nosystematics(ipred, imeas);
    }
    return chi2;
}

void InputFunction::get_hessian(Matrix& hess, const ObservableVector& obs)
{
    assert(int(obs.size()) == nobservables());
    hess = boost::numeric::ublas::zero_matrix<double>(nobservables(),
                                                      nobservables());

    for(int i=0; i<int(lcv_.size()); ++i) {
        if(!is_constraint_[i]) {
            ObservableVector iobs(lcv_[i]->n_claimed_observables());
            Matrix ihess(lcv_[i]->n_claimed_observables(),
                         lcv_[i]->n_claimed_observables());
            for(int j=0; j<int(iobs.size()); ++j)
                iobs[j] = obs[lcv_[i]->external_index(j)];
            lcv_[i]->get_hessian(ihess, obs);
            for(int j=0; j<int(iobs.size()); ++j)
                for(int k=0; k<int(iobs.size()); ++k)
                    hess(lcv_[i]->external_index(j),
                         lcv_[i]->external_index(k)) += ihess(j,k);
        }
    }
}


}
