/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "model.hpp"

#include <iostream>

#include <boost/numeric/ublas/matrix_proxy.hpp>

#include <boost/random/uniform_real.hpp>
#include <boost/random/variate_generator.hpp>

#include "comparison.hpp"


namespace myfitter {

typedef boost::numeric::ublas::matrix_row<Matrix> MatrixRow;
typedef boost::numeric::ublas::zero_vector<double> ZeroVector;


// class Model

int Model::numerical_derivative_(int i)
{
    assert(i >= 0 && i < nparameters());
    int status;
    if(needs_update_)
        if((status=calc())) return status;
    MatrixRow rowi(derivatives_, i);
    if(fixed(i)) {
        rowi = ZeroVector(nobservables());
        return 0;
    }
    ObservableVector obs0(observables_);
    double oldpari = par_val_[i];
    double dpari = derivative_epsilon()*scale(i);
    par_val_[i] += dpari;
    needs_update_ = true;
    if((status=calc())) return status;
    rowi = (observables_ - obs0)*(1./dpari);
    observables_ = obs0;
    par_val_[i] = oldpari;
    return 0;
}

int Model::scan(int n)
{
    boost::variate_generator<boost::mt19937&, boost::uniform_real<> >
        rng(base_rng_, boost::uniform_real<>(0., 1.));

    ParameterVector smin(nparameters());
    ParameterVector smax(nparameters());
    for(int i=0; i<nparameters(); ++i) {
        smin[i] = scan_min_[i];
        smax[i] = scan_max_[i];
        if(has_lower_limit(i)) smin[i] = std::max(smin[i], lower_limit(i));
        if(has_upper_limit(i)) smax[i] = std::min(smax[i], upper_limit(i));
    }

    double x;
    int status;
    for(int i=0; i<n; ++i) {
        for(int j=0; j<nparameters(); ++j) {
            if(!fixed(j)) {
                x = rng();
                parameter(j, (1.-x)*smin[j] + x*smax[j]);
            }
        }
        if((status=store()) != 0) return status;
    }
    return 0;
}


int Model::store()
{
    int status;
    if(needs_update_)
        if((status=calc()) != 0) return status;
    par_map_.push_back(parameters());
    obs_map_.push_back(observables());
    return 0;
}

bool Model::check_derivatives(double rprec, double aprec) 
{
    int status;
    for(int i=0; i<nparameters(); ++i)
        if((status=numerical_derivative_(i))) {
            std::cerr << "calculation of numerical derivatives with respect to\
parameter " << i << "failed with exit status " << status << std::endl;
            return false;
        }
    
    Matrix nderivs(derivatives_);
    if((status = calc_deriv())) {
        std::cerr << "calculation of derivatives failed with exit status "
                  << status << std::endl;
        return false;
    }

    bool ok = true;
    for(int i=0; i<nparameters(); ++i)
        for(int j=0; j<nobservables(); ++j)
            if(differ(derivatives_(i,j), nderivs(i,j), rprec, aprec)) {
                std::cout << "discrepancy for parameter " << i
                          << ", observable " << j
                          << ":\nnumerical: " << nderivs(i,j)
                          << "  custom:     " << derivatives_(i,j)
                          << std::endl;
                ok = false;
            }
    return ok;
}


}
